

#include <stdio.h>
#include <assert.h>

#include <set>
#include <algorithm>


#include "ndfa_node.h"
#include "utils.h"



/****** NDFA_Node public methods ******/

/** NDFA_Node constructor
 *
 * A new node for an NDFA is produced. It is neither an accept nor
 * reject state, and has no transitions. The bit number to be tested
 * for the transitions is given as a parameter to the constructor.
 */
NDFA_Node :: NDFA_Node ( int test_bit_number )
{
    assert (( test_bit_number >= 0 )
            && ( test_bit_number < ( BITS_PER_OPCODE + 1 ))) ;

    accept_state = 0L ;
    is_accept_state = false ;
    is_reject_state = false ;
    this -> test_bit_number = test_bit_number ;

    for ( int ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        transitions [ ts ] . clear () ;
    }

    is_deterministic = false ;

    visits = 0 ;
}

/** NDFA_Node destructor
 *
 * The NDFA node is deleted along with all its children. All NDFA-specific
 * memory is freed:- no attempt is made to free the accept state memory if
 * any.
 */
NDFA_Node :: ~NDFA_Node ()
{
    Delete_Children () ;
}



/** Add_Transition
 *
 * Adds a transition from this node to a new one. The new node is
 * returned. Parameters specify whether the transition is on '0',
 * '1', or both.
 */
NDFA_Node * NDFA_Node :: Add_Transition ( bool transition_on_zero ,
                                    bool transition_on_one )
{
    assert ( transition_on_zero || transition_on_one ) ;

    NDFA_Node * nn = new NDFA_Node ( test_bit_number + 1 ) ;

    if ( transition_on_zero )
    {
        transitions [ 0 ] . insert ( nn ) ;
    }
    if ( transition_on_one )
    {
        transitions [ 1 ] . insert ( nn ) ;
    }
    
    Check_For_Determinism () ;
    return nn ;
}

/** Make_Reject_State
 *
 * Makes this node a reject state.
 */
void NDFA_Node :: Make_Reject_State ()
{
    is_reject_state = true ;
    is_accept_state = false ;
    Delete_Children () ;
    Check_For_Determinism () ;
}

/** Make_Accept_State
 *
 * Makes this node an accept state.
 * If this node is already an accept state, the new accept state
 * information must be the same as the old. If it is not, an exception
 * (SharedAcceptStateException) will be thrown.
 * If this node is already a reject state, nothing happens. (Reject
 * takes priority over accept).
 */
void NDFA_Node :: Make_Accept_State ( Accept_State * accept_info )
{
    if ( is_reject_state )
    {
        return ;
    }

    if ( is_accept_state )
    {
        if ( accept_state != accept_info )
        {
            assert ( accept_state != 0L ) ;
            assert ( accept_info != 0L ) ;

            /* An NDFA node cannot possibly be two different types of
            accept states. This would mean that two opcodes had the
            same bit pattern but a different meaning. */
            throw new SharedAcceptStateException (
                        accept_state , accept_info ) ;
        }
    }
    is_accept_state = true ;
    is_reject_state = false ;
    accept_state = accept_info ;

    Check_For_Determinism () ;
}


/** Merge_In
 * 
 * The provided NDFA is copied into this one. A deep copy is made of all 
 * nodes in the 2nd NDFA, meaning it can be freely deleted. However, 
 * shallow copies are made of any accept_state records.
 */
void NDFA_Node :: Merge_In ( NDFA_Node * source_root ,
                                    bool copy_reject_states )
{
    NDFA_Node   * new_ndfa_node ;

    /* Copy an accept state from source to this, checking to ensure
     * that an existing accept state is not overwritten. */
    if ( source_root -> is_accept_state )
    {
        Make_Accept_State ( source_root -> accept_state ) ;
    }

    /* Copy all transitions from source to target */
    for ( int ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        NDFA_Node_Set_Iterator  iter ;
        NDFA_Node *             item ;

        iter = source_root -> transitions [ ts ] . begin () ;

        while ( iter != source_root -> transitions [ ts ] . end () )
        {
            item = (* iter) ;

            new_ndfa_node = new NDFA_Node 
                                ( source_root -> test_bit_number + 1 ) ;

            new_ndfa_node -> Merge_In ( item , copy_reject_states ) ;
            
            this -> transitions [ ts ] . insert ( new_ndfa_node ) ;

            iter ++ ;
        }
    }

    if (( source_root -> is_reject_state )
    && ( copy_reject_states ))
    {
        this -> is_reject_state = true ;
        this -> is_accept_state = false ;
    }

    Check_For_Determinism () ;
}






/** Delete_Dead_Branches
 *
 * A dead branch is one with either:
 * - no _enabled_ accept states amongst any of its descendants.  
 * - a zero visit count (indicating that it will never be used in practice)
 * This function tests to see if "this"
 * is the parent of any dead branches.  Any descendant dead branches are 
 * removed and deallocated. 
 *
 * It returns TRUE if children were removed.
 */
bool NDFA_Node :: Delete_Dead_Branches ()
{
    NDFA_Node_Set_Iterator  iter ;
    NDFA_Node *             item ;
    bool                    is_dead_branch = false ;
    bool                    no_children = true ;

    /* First, eliminate dead branches from the children */
    for ( int ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        iter = this -> transitions [ ts ] . begin () ;
        while ( iter != this -> transitions [ ts ] . end () )
        {
            item = (* iter) ;

            if ( item -> Delete_Dead_Branches () )
            {
                iter ++ ;
                Delete_Item_Before ( & this -> transitions [ ts ] , 
                                        iter ) ;
            } else {
                iter ++ ;
            }
        }

        if ( ! this -> transitions [ ts ] . empty () )
        {
            no_children = false ;
        }
    }

    /* This is a dead branch if both the branches off it are NULL
    (no children - i.e. all_sub_branches_dead == true) and either :
        1. it is not an accept state 
     or 2. it is an accept state, but the accept state has been disabled
     or 3. the visit counter is zero (in which case, the same will have been
            true for the children, so there won't be any children)

    Return false iff this is a dead branch.*/

    if ( no_children )
    {
        if ( visits == 0 )
        {
            is_dead_branch = true ;
        }
        if ( this -> is_accept_state )
        {
            if ( ! this -> accept_state -> Is_Enabled () )
            {
                is_dead_branch = true ;
            }
        } else {
            is_dead_branch = true ;
        }
    }

    return is_dead_branch ;
}



/** Compress
 *
 * Several algorithms are applied recursively to the NDFA tree to reduce the
 * number of nodes while maintaining equivalence.
 *
 * 1. Nodes with one child are replaced by the child.
 * 2. Nodes with two children are replaced by either child if the
 *    children are clearly equivalent (both the same accept state,
 *    or both reject states)
 * 
 * These are surprisingly effective. Unfortunately, once compressed,
 * the tree cannot be converted to a DFA again, because
 * the Make_Deterministic algorithm is unable to handle the idea that
 * nodes at the same tree depth might be testing different bits.
 *
 * The tree must be a DFA before it is compressed.
 */
void NDFA_Node :: Compress ()
{
    NDFA_Node *             compressed_out_node ;
    NDFA_Node_Set_Iterator  iter ;
    NDFA_Node *             item ;
    int                     ts ;

    assert ( is_deterministic ) ;

    MESSAGE4 ( "Compress() begins bit = %d\n" , test_bit_number ) ;

    /* Recurse down and do the children if any. */
    for ( ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        iter = this -> transitions [ ts ] . begin () ;
        while ( iter != this -> transitions [ ts ] . end () )
        {
            item = (* iter) ;

            /* Is this child a reject state? If so, it can be removed. */
            if ( item -> is_reject_state )
            {
                /* a quick sanity check assures that the tree is
                 * sane - no state is a reject and accept state */
                assert ( ( ! item -> is_accept_state )
                    && item -> transitions [ 0 ] . empty ()
                    && item -> transitions [ 1 ] . empty () ) ;
                iter ++ ;
                Delete_Item_Before ( & this -> transitions [ ts ] , 
                                        iter ) ;
                MESSAGE4 ( "Compress() removed reject state bit = %d\n" , 
                                            test_bit_number ) ;
            } else {
                /* Attempt to compress it */
                item -> Compress () ;

                iter ++ ;
            }
        }
    }

    NDFA_Node_Set * tset0 = & ( this -> transitions [ 0 ] ) ;
    NDFA_Node_Set * tset1 = & ( this -> transitions [ 1 ] ) ;

    /* Compressions are possible if a node has only one child. */
    if ( tset1 -> empty ()
    && ( tset0 -> size () == 1 ))
    {
        MESSAGE4 ( "Compress() node compressed out (type 1) bit = %d\n" , 
                                    test_bit_number ) ;

        compressed_out_node = DFA_Transition ( 0 ) ;
    } else if ( tset0 -> empty ()
    && ( tset1 -> size () == 1 ))
    {
        MESSAGE4 ( "Compress() node compressed out (type 2) bit = %d\n" , 
                                    test_bit_number ) ;

        compressed_out_node = DFA_Transition ( 1 ) ;
    /* or if there are two children, but they are both
    accept states of the same type. */
    } else if ( (( tset1 -> size () == 1 )
    && ( tset0 -> size () == 1 )
    && ( DFA_Transition ( 1 ) -> is_accept_state )
    && ( DFA_Transition ( 0 ) -> is_accept_state )
    && ( DFA_Transition ( 1 ) -> accept_state ==
                        DFA_Transition ( 0 ) -> accept_state )
    && ( DFA_Transition ( 1 ) -> test_bit_number ==
                        DFA_Transition ( 0 ) -> test_bit_number ))
    /* or if there are two children, but they are both reject states. */
    || (( tset1 -> size () == 1 )
    && ( tset0 -> size () == 1 )
    && ( DFA_Transition ( 1 ) -> is_reject_state )
    && ( DFA_Transition ( 0 ) -> is_reject_state )) )
    {
        MESSAGE4 ( "Compress() node compressed out (type 3) bit = %d\n" , 
                                    test_bit_number ) ;

        compressed_out_node = DFA_Transition ( 1 ) ;
                /* or tset0.. doesn't matter, they're the same. */

        if ( compressed_out_node != DFA_Transition ( 0 ) )
        {
            /* The two nodes are equivalent but are at different
             * memory locations, so the 0 node must be deleted.
             * (It will soon be inaccessible). */
            delete DFA_Transition ( 0 ) ;
        }
    } else {
        compressed_out_node = 0L ;
    }

    if ( compressed_out_node != 0L )
    {
        /* Move the grandchildren up one generation.
        The node we are looking at, this, Takes on the
        properties of the compressed_out_node. */
        Replace_With ( compressed_out_node ) ;
    }
    MESSAGE4 ( "Compress() ends bit = %d\n" , test_bit_number ) ;
}


/** Print_NDFA_DAG
 *
 * The NDFA tree is drawn out in ASCII format to the specified device.
 */
void NDFA_Node :: Print_NDFA_DAG ( bool with_pointers , FILE * fd ,
                                        const char * old_tab_str )
{
    if ( is_accept_state )
    {
        assert ( accept_state != 0L ) ;

        accept_state -> Print_Info ( fd , old_tab_str ) ;
        if ( with_pointers )
        {
            fprintf ( fd , "%s     ptr=%p\n" , old_tab_str , this ) ;
        }
    } else if ( is_reject_state )
    {
        if ( with_pointers )
        {
            fprintf ( fd , "%s   REJECT ptr=%p\n" , old_tab_str , this ) ;
        } else {
            fprintf ( fd , "%s   REJECT\n" , old_tab_str ) ;
        }
    } else {
        NDFA_Node_Set_Iterator  iter ;
        NDFA_Node *             item ;
        char *                  tab_str = 
                                    new char [ strlen ( old_tab_str ) + 4 ] ;


        strcpy ( tab_str , old_tab_str ) ;
        strcat ( tab_str , "1  " ) ;
        fprintf ( fd , "%s+-ir(%d)=one:\n" , tab_str , 
                BITS_PER_OPCODE - 1 - test_bit_number ) ;

        iter = transitions [ 1 ] . begin () ;
        while ( iter != transitions [ 1 ] . end () )
        {
            item = (* iter) ;

            item -> Print_NDFA_DAG ( with_pointers , fd , tab_str ) ;
            iter ++ ;
        }
        fprintf ( fd , "%s\n" , tab_str ) ;


        strcpy ( tab_str , old_tab_str ) ;
        strcat ( tab_str , "0  " ) ;
        fprintf ( fd , "%s+-ir(%d)=zero:\n" , tab_str , 
                BITS_PER_OPCODE - 1 - test_bit_number ) ;

        iter = transitions [ 0 ] . begin () ;
        while ( iter != transitions [ 0 ] . end () )
        {
            item = (* iter) ;

            item -> Print_NDFA_DAG ( with_pointers , fd , tab_str ) ;
            iter ++ ;
        }
        fprintf ( fd , "%s\n" , tab_str ) ;


        delete [] tab_str ;
    }
}

/** Get_Size
 *
 * Returns the number of nodes in the tree.
 */
int NDFA_Node :: Get_Size ()
{
    NDFA_Node_Set_Iterator  iter ;
    NDFA_Node *             item ;
    int                     count = 1 ;

    if (( ! is_reject_state )
    && ( ! is_accept_state ))
    {
        for ( int ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
        {
            iter = transitions [ ts ] . begin () ;
            while ( iter != transitions [ ts ] . end () )
            {
                item = (* iter) ;

                count += item -> Get_Size () ;
                iter ++ ;
            }
        }
    }

    return count ;
}

/** Make_Deterministic
 *
 * The NDFA is converted to a DFA. A depth-bounded depth first tree
 * algorithm is used - Transform_NDFA_To_DFA is the procedure that
 * actually does the work. This algorithm is used because it avoids
 * the possibility that a very long branch of the NDFA may be uselessly
 * converted to a DFA before it is realised that much of it will be
 * rejected.
 *
 * Shallow copies of accept state pointers in the NDFA are made. Checking
 * is done to ensure there are no ambiguous accept states. If a state is
 * both an accept and a reject state, it is assumed to be a reject state.
 *
 * The return value is the number of nodes in the DFA.
 *
 * This will not work correctly on all compressed trees - an assertion
 * will fail if the tree has been compressed. See the comment for
 * Compress().
 */
void NDFA_Node :: Make_Deterministic ()
{
    MESSAGE4 ( "Make_Deterministic ()\n" ) ;

    /* Create a new root for the DFA */
    NDFA_Node * dfa_root = new NDFA_Node ( test_bit_number ) ;

    /* Transform the NDFA into a DFA. This is a depth-bounded
     * depth first tree operation - the depth is increased until
     * the operation finishes. */
    for ( int i = test_bit_number ; i <= BITS_PER_OPCODE ; i ++ )
    {
        Transform_NDFA_To_DFA ( dfa_root , i ) ;
    }

    /* Remove children of the NDFA tree root */
    Delete_Children () ;

    /* Destroy the NDFA tree, replacing it with the DFA tree. */
    Replace_With ( dfa_root ) ;

    Check_For_Determinism () ;
    assert ( is_deterministic ) ;
}



/** Get_Accept_State
 * 
 * Returns the accept state information for the node (if any) or 0.
 */
Accept_State * NDFA_Node :: Get_Accept_State ()
{
    if ( is_accept_state )
    {
        return accept_state ;
    } else {
        return 0L ;
    }
}


/** DFA_Transition
 *
 * Returns the node reached by a transition ts. It returns 0L if
 * no node is reached.
 */
NDFA_Node * NDFA_Node :: DFA_Transition ( int ts )
{
    NDFA_Node_Set_Iterator  iter ;

    if ( transitions [ ts ] . empty () )
    {
        return 0L ;
    } else {
        iter = transitions [ ts ] . begin () ;
        return (* iter) ;
    }
}

/****** NDFA_Node private methods ******/


/** Delete_Item_Before
 *
 * This convenience procedure finds the NDFA node before the current item
 * in "set" (i.e. "iter" - 1). This NDFA node is removed from the set and
 * deleted.
 */
void NDFA_Node :: Delete_Item_Before ( NDFA_Node_Set * set , 
                                        NDFA_Node_Set_Iterator iter )
{
    NDFA_Node *             item ;

    iter -- ;
    item = (* iter) ;

    delete item ;

    set -> erase ( iter ) ;
}





/** Check_For_Determinism
 *
 * Examines the current node (no recursion) and its children
 * and sets the deterministic flags appropriately.
 *
 * Node is deterministic if there are 0 or 1 children
 * for each type of transition and those children are deterministic too.
 */
void NDFA_Node :: Check_For_Determinism ()
{
    is_deterministic = true ; 
    for ( int ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        if ( transitions [ ts ] . size () > 1 )
        {
            // not deterministic - 2 or more children
            is_deterministic = false ; 
        } else if ( ! transitions [ ts ] . empty () )
        {
            if ( ! DFA_Transition ( ts ) -> is_deterministic )
            {
                // child is not deterministic
                is_deterministic = false ; 
            }
        }
    }
}




/** Delete_Children
 *
 * Used for the conversion of a node into a rejecting node, or for
 * freeing the data structures.
 */
void NDFA_Node :: Delete_Children ()
{
    NDFA_Node_Set_Iterator  remove_iter ;

    /* Unfortunately a child node may be in both ndfa_root -> transitions[1]
    and ndfa_root -> transitions[0]. So items from transitions[0] are moved
    to transitions[1] before deletion */
    remove_iter = transitions [ 0 ] . begin () ;
    while ( remove_iter != transitions [ 0 ] . end () )
    {
        NDFA_Node * n = (* remove_iter) ;

        transitions [ 1 ] . insert ( n ) ;
        remove_iter ++ ;
    }
        
    remove_iter = transitions [ 1 ] . begin () ;
    while ( remove_iter != transitions [ 1 ] . end () )
    {
        NDFA_Node * n = (* remove_iter) ;

        delete n ;

        remove_iter ++ ;
    }
    transitions [ 0 ] . clear () ;
    transitions [ 1 ] . clear () ;
}


/** Transform_NDFA_To_DFA
 *
 * Does the work for Make_Deterministic(). The NDFA (this) is
 * converted to a DFA (dfa_root) by a depth bounded depth first
 * tree operation.
 *
 * This will not work correctly on all compressed trees - an assertion
 * will fail if the tree has been compressed. See the comment for
 * Compress().
 */
void NDFA_Node :: Transform_NDFA_To_DFA ( NDFA_Node * dfa_root ,
                                            int depth_remaining )
{
    NDFA_Node_Set_Iterator  iter ;
    NDFA_Node *             item ;
    int                     ts ;
 
    if ( depth_remaining < 0 )
    {
        /* Implement the depth bound */
        return ;
    }

    if (( dfa_root -> is_reject_state )
    || ( dfa_root -> is_accept_state ))
    {
        /* The DFA node is an accept/reject state. The children of
         * the node, if any, are not important. */
        return ;
    }

    /* If any NDFA nodes are in each transition set, check to 
     * see if any of them are accept or reject states... */
    for ( ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        if ( ! this -> transitions [ ts ] . empty () )
        {
            NDFA_Node * dfa_child = dfa_root -> DFA_Transition ( ts ) ;

            if ( dfa_child == 0L )
            {
                dfa_child = new NDFA_Node 
                                ( dfa_root -> test_bit_number + 1 ) ;
                dfa_root -> transitions [ ts ] . insert ( dfa_child ) ;
            }


            iter = this -> transitions [ ts ] . begin () ;
            while ( iter != this -> transitions [ ts ] . end () )
            {
                item = (* iter) ;
                if ( item -> is_reject_state )
                {
                    /* The NDFA node is a reject state. Go no further: the
                    DFA node must also be a reject state. */
                    dfa_child -> Make_Reject_State () ;
                } else if ( item -> is_accept_state )
                {
                    /* The NDFA node is an accept state. Ensure that, if
                     * the DFA node is already known to be an accept state,
                     * that it is the SAME accept state. Otherwise,
                     * it is non-deterministic and there is nothing
                     * we can do to change that. */

                    dfa_child -> Make_Accept_State ( item -> accept_state ) ;
                } else {
                    /* Regular transition node.
                     * Ensure that all children reached by transition ts in 
                     * the NDFA are testing the same bit number. */
                     assert ( dfa_child -> test_bit_number ==
                                        item -> test_bit_number ) ;
                }
                iter ++ ;
            }
        }
    }

    if (( dfa_root -> is_reject_state )
    || ( dfa_root -> is_accept_state ))
    {
        /* The DFA node has become, as a result of the nodes just added
         * to it, an accept/reject state. The children of
         * the node, if any, are not important. */
        return ;
    }

    /* Transform the child subtrees */
    for ( ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        NDFA_Node * dfa_child = dfa_root -> DFA_Transition ( ts ) ;

        if ( dfa_child != 0L )
        {
            iter = this -> transitions [ ts ] . begin () ;
            while ( iter != this -> transitions [ ts ] . end () )
            {
                item = (* iter) ;

                item -> Transform_NDFA_To_DFA ( dfa_child ,
                                    depth_remaining - 1 ) ;
                iter ++ ;
            }
        }
    }
    dfa_root -> Check_For_Determinism () ;
}

/** Replace_With
 *
 * The important fields of n are copied to this. Then, n is deleted.
 * The aim is to allow the node <this> to be replaced by n, so that
 * n is moved up the tree.
 */
void NDFA_Node :: Replace_With ( NDFA_Node * n )
{
    this -> accept_state = n -> accept_state ;
    this -> is_accept_state = n -> is_accept_state ;
    this -> is_reject_state = n -> is_reject_state ;
    this -> test_bit_number = n -> test_bit_number ;

    for ( int ts = 0 ; ts < TRANSITION_TYPES ; ts ++ )
    {
        this -> transitions [ ts ] = n -> transitions [ ts ] ;
        n -> transitions [ ts ] . clear () ;
    }

    delete n ;
}


